import logging
import os
import re
from typing import Any

from box import Box
from dotenv import find_dotenv, load_dotenv
from tomlkit import parse, items

logger = logging.getLogger(__name__)

try:
    import boto3
    from botocore.exceptions import ClientError
    _boto_available = True
except ImportError:
    logger.warning("boto3 is not installed - install it to use SSM parameters.")
    _boto_available = False


class Settings:
    """Settings class that can be accessed using either dict notation (settings.get('abc')) or
    dot notation (settings.snowflake.password). Reads from toml file and requires a table/section
    called 'default', along with a table/section for each environment, e.g. [local], [dev], [prod].
    Supports nested tables with dot notation, e.g. [local.snowflake].
    Inspired by ConfigParser and DynaConf.
    """

    TOML_TO_BUILTIN_MAP = {
        items.String: str,
        items.Bool: bool,
        items.Integer: int,
        items.Float: float,
    }

    INTERNAL_ATTRS = ["_store", "_ssm", "env"]

    def __init__(self, config_filepath: str = None, env: str = os.getenv("PROJECT_ENV", "local")):
        self._store = Box()
        self.env = env
        self._ssm = boto3.client("ssm") if _boto_available else None

        # Load from .env file if exists. Will set env variables for use in .ini files.
        load_dotenv(find_dotenv(usecwd=True), verbose=True)

        if config_filepath:
            self.load(config_filepath)

    def load(self, config_filepath: str):
        if not os.path.exists(config_filepath):
            raise OSError("Could not load the default or provided settings file.")

        self.clear()

        with open(config_filepath, "r") as f:
            settings_data = parse(f.read())

        if "default" not in settings_data:
            raise Exception("Settings file missing required section 'default'")

        for table, items in settings_data.items():
            if table.startswith(self.env) or table == "default":
                for k, v in items.items():
                    self._set_value_from_config(k, v)

    def clear(self):
        self._store = Box()

    def _set_value_from_config(self, name: str, value: Any, parent: str = None):
        # If an env var exists, it takes precedence over everything else
        if name.upper() in os.environ and not parent:
            self.set_attr(name, os.getenv(name.upper()), parent)
        # Or if there's an env var for nested child attr, such as PARENT_CHILD, it takes precedence over orig value
        elif parent and f"{parent.upper()}_{name.upper()}" in os.environ:
            self._set_from_parent_env_var(name, value, parent)
        # If the value is a dict, recursively call this fcn to un-nest it's fields and set them.
        elif isinstance(value, dict):
            for k, v in value.items():
                self._set_value_from_config(k, v, name)
        # If the value defined in toml is "${MY_VAR}", it must be set from env var. Check it exists and set it.
        elif isinstance(value, str) and value.startswith("${") and value.endswith("}"):
            self._set_from_env_var_interpolation(name, value, parent)
        # If value defined with "ssm:" prefix, try fetching it from SSM parameter store.
        elif isinstance(value, str) and value.startswith("ssm:"):
            self._set_from_ssm(name, value, parent)
        # Otherwise, it should be standard type that can be set directly.
        else:
            self.set_attr(name, value, parent)

    def _set_from_env_var_interpolation(self, name: str, value: Any, parent: str = None):
        """Expecting environment variable with the value between ${}. If not found, variable is set to None."""
        var_name = re.findall(r'\${(.*?)}', value)[0]

        if var_name in os.environ:
            self.set_attr(name, os.getenv(var_name), parent)
        else:
            self.set_attr(name, None, parent)

    def _set_from_parent_env_var(self, name: str, value: Any, parent: str = None):
        """Allows overriding nested attributes by setting env var as PARENT_CHILD."""
        attr = os.getenv(f"{parent.upper()}_{name.upper()}")
        try:
            if type(value) in self.TOML_TO_BUILTIN_MAP:
                attr = self.TOML_TO_BUILTIN_MAP[type(value)](attr)
            elif type(value) == bool:
                attr = attr.lower() == "true"
            else:
                attr = type(value)(attr)
        except ValueError:
            logger.info(f"Could not cast setting {parent}.{name} with value {attr} to type {type(value)}")
            pass
        self.set_attr(name, attr, parent)

    def _set_from_ssm(self, name: str, value: Any, parent: str = None):
        """Gets parameter from ssm, where value starts with `ssm:` """
        if not _boto_available:
            raise ImportError("boto3 not available and is required to read from SSM.")

        try:
            param = self._ssm.get_parameter(
                Name=value.replace("ssm:", ""),
                WithDecryption=True
            )

            if param.get("Parameter"):
                self.set_attr(name, param["Parameter"]["Value"], parent)
        except ClientError as e:
            # Best effort to load parameter
            logger.error(e)

    def set_attr(self, name: str, value: Any, parent: str = None):
        if type(value) in self.TOML_TO_BUILTIN_MAP:
            value = self.TOML_TO_BUILTIN_MAP[type(value)](value)

        if isinstance(value, dict):
            value = Box(value)

        if parent:
            if parent not in self._store:
                self._store[parent] = Box()

            self._store[parent][name] = value
        else:
            self._store[name] = value

    def __dir__(self):
        """Enable auto-complete for code editors"""
        return (
            self.INTERNAL_ATTRS
            + [k.lower() for k in self._store.keys()]
        )

    def __getattr__(self, name):
        """Allow getting keys from self._store using dot notation"""
        if name in self.INTERNAL_ATTRS:
            return super(Settings, self).__getattribute__(name)
        value = getattr(self._store, name)
        return value

    def __setattr__(self, name, value):
        """Allow `settings.FOO = 'value'` while keeping internal attrs."""
        if name in self.INTERNAL_ATTRS:
            super(Settings, self).__setattr__(name, value)
        else:
            self.set_attr(name, value)

    def __contains__(self, item):
        """Respond to `item in settings`"""
        return item.upper() in self._store or item.lower() in self._store

    def __getitem__(self, item):
        """Allow getting variables as dict keys `settings['KEY']`"""
        value = self._store.get(item)
        if value is None:
            raise KeyError(f"{item} does not exist")
        return value

    def __setitem__(self, key, value):
        """Allow `settings['KEY'] = 'value'`"""
        self.set_attr(key, value)

    def __iter__(self):
        """Redirects to store object"""
        yield from self._store

    def items(self):
        """Redirects to store object"""
        return self._store.items()

    def keys(self):
        """Redirects to store object"""
        return self._store.keys()

    def values(self):
        """Redirects to store object"""
        return self._store.values()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['configur']

package_data = \
{'': ['*']}

install_requires = \
['python-box>=5.3.0,<6.0.0',
 'python-dotenv>=0.18.0,<0.19.0',
 'tomlkit>=0.7.2,<0.8.0']

setup_kwargs = {
    'name': 'configur',
    'version': '0.1.6',
    'description': 'TOML-based configuration for Python',
    'long_description': '## Configur\n#### TOML-based configuration for Python\n\n### Features\n* TOML file configuration definition (supports nested tables, multiple environments)\n* Fetching variables from AWS Parameter Store (SSM), secure or standard\n* Override any configuration at runtime with an environment variable\n* Simple logging config: easily set levels per package, special handling for AWS Lambda\n\n\n#### Configuration\nAll configuration is defined in a single TOML file, e.g., `settings.toml` or `config.toml`. See\nthe [example](https://github.com/StudioTrackr/configur/blob/main/example.toml) for a better idea of how it\'s structured.\n\nCertain secrets (e.g. passwords) that should not be checked into Git can be interpolated from\nenvironment variables. If you declared an item as `my_var = "${EXAMPLE_VALUE}"`, you will\nneed to have an environment variable `EXAMPLE_VALUE` defined. Configur also come with `dotenv`, so feel free\nto define a `.env` file for storing these environment variables.\n\nTo avoid passing secrets in plaintext in the run environment, you can also define variables in AWS Systems Manager\nParameter Store (SSM). To do so, simply prefix the value with `ssm:` in the TOML file, and the code will automatically\nfetch and decode the param at runtime, assuming the proper credentials are available to get and decrypt the parameter.\n\nThese configuration files are parsed by the [TOML Kit](https://github.com/sdispater/tomlkit) library, and then stored \nas attributes on an object called [Settings](#Settings). \n\nEvery settings.toml file must have a `default` table. These default values are shared across all environments.\n\n```\n[default]\noutput_file_name = "example.csv"\n```\n\nAfter the `default` table, a table for each environment should exist, e.g. "local", "dev", and "prod". These tables override\nany values that exist in the `default` table with the same name; in other words - environment settings take precedence over default settings.\nBelow we have two environments, `local` and `dev`, each with a custom `database_username` variable. In `dev`, we override the default `output_file_name`.\n```\n[default]\noutput_file_name = "example.csv"\n\n[local]\ndatabase_username = "local_user"\n\n[dev]\noutput_file_name = "example_dev.csv"\ndatabase_username = "dev_user"\n```\n\nSince TOML and TOML Kit support nested tables/sections, we use them within the context of an environment. To declare a group of settings\nfor a given environment, prefix the table name with the name of the environment, e.g. `local.mysql`.\n\n```\n[default]\noutput_file_name = "example.csv"\n\n[local]\n    [local.mysql]\n    username = "local_user"\n    password = "secret password"\n```\n\nRead more below for how these values are accessed in Settings.\n\n#### Settings\nThe [Settings](https://github.com/StudioTrackr/configur/blob/main/configur/config.py) class reads `settings.toml` files, and sets attributes on itself for \nevery item in the config file. However, it will only ever load settings from the `default` table and the table (and its children) matching \nthe current environment, e.g., `local`. The environment **must** be set with a variable `PROJECT_ENV`, otherwise the fallback \nis `local` so nothing will ever touch production by accident. The value of this variable needs to match a corresponding section in your `settings.toml` file,\nbut you are free to name environments as you wish. There\'s no difference if you call an environment `dev` or `sandbox` or `test`, so long as\nyou set `PROJECT_ENV=dev` or `PROJECT_ENV=sandbox` or `PROJECT_ENV=test`.\n\nInitialization:\n```\nfrom configur import Settings\nsettings = Settings(config_filepath=custom_path + "settings.toml")\n```\n\nBy default, it extracts the environment from the `PROJECT_ENV` variable, but you can override if needed by passing an `env` argument:\n```\nsettings = Settings(config_filepath=custom_path + "settings.toml", env="dev")\n```\n\nAfter your config file is read, the default and environment-specific values are set to be accessible from the Settings object\nin either dict-notation, or dot-notation. Inspired by [Dynaconf](https://www.dynaconf.com/), this means you can do the following:\n```\n# Get\nsettings.user\nsettings.get("user")\nsettings["user"]\n\n# Get nested settings e.g. from [local.mysql] section of settings.toml\nsettings.mysql.user\nsettings.get("mysql").user\nsettings["mysql"]["user"]\n\n# Set\nsettings.user = "admin"\nsettings["user"] = "admin"\n\n# Iterate\nfor x in settings.items()\nfor x in settings.keys()\nfor x in settings.values()\n```\n\nThis flexibility makes it easier to access settings vs always having to use dict-notation, get environment variables every time, \nor use ConfigParser and pass the section for every variable. \n\nBecause Settings is a dict-like object, you can also set values to update config or store state as your job progresses. \nThis can however introduce side effects since you are bringing state into functions, but it can be handy to throw variables in here\ninstead of passing them down a large tree of functions as standard arguments.\n\n#### Logging\nConfigur provides a function `init_logging` which initializes handlers, formatters, and loggers with Python\'s logging.config.dictConfig.\nSimilar to Settings, you can pass a custom `env` argument but the default is your `PROJECT_ENV` environment variable.\nWe set the root log level based on your environment (local, dev = DEBUG, all others = INFO), but you can also pass this as an\noverride with the `root_level` argument, OR by setting an environment variable `ROOT_LOG_LEVEL`. We also expose the logging.config option\n`disable_existing_loggers` as an argument, defaulted to False. \n\n##### Customizing module loggers: \nYou can customize the log level for any module by passing a dictionary where keys are module names, and values are log levels. \nExample as read from `settings.toml`:\n\n```\n# settings.toml\n[logging]\nboto3 = "DEBUG"\nbotocore = "ERROR"\n\n# main.py\ninit_logging(loggers=settings.logging)\n```\n\n##### Special Handling for AWS Lambda\nSince AWS Lambda controls the logging environment, we can\'t/shouldn\'t set any custom formatters or logging config. What \nwe can do though is set the overall log level. When running a lambda recipe, use the `is_lambda` option, which when set to True\nwill skip the dictConfig initialization and just call `logging.getLogger().setLevel(log_level)` with either the default\nenvironment level, or a custom level passed in like `init_logging(level="DEBUG")`.\n',
    'author': 'Studio Trackr',
    'author_email': 'info@studiotrackr.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/StudioTrackr/configur',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
